from textwrap import dedent


def inline_code(code_func: str) -> str:
    code_wrapper = """\
        def solution_wrapper():

            {code_func}

            return solution
        """
    # ко второй строке и далее долбавляем один таб
    code_func_lines = dedent(code_func).split('\n')
    code_wrapper_lines = dedent(code_wrapper).split('\n')
    res_lines: list[str] = []
    for line in code_wrapper_lines:
        if line.strip() == "{code_func}":
            for func_line in code_func_lines:
                res_lines.append("\t" + func_line)
        else:
            res_lines.append(line)

    return "\n".join(res_lines)



import os

from analyst_klondike.features.code_import.analysis.code_analysis import (
    create_quiz_from_code
)
from analyst_klondike.features.code_import.ui_actions.actions import ImportTasksAction
from analyst_klondike.state.app_dispatch import app_dispatch


def import_task_action(quiz_id: str, file_path: str | None) -> None:

    if file_path is None or not os.path.exists(file_path):
        raise FileNotFoundError(f"{file_path} not exists")

    with open(file_path, 'r', encoding='UTF-8') as f:
        code = f.read()
        python_tasks = create_quiz_from_code(quiz_id, code)
        app_dispatch(ImportTasksAction(quiz_id, python_tasks))



from analyst_klondike.features.data_context.init_action import (
    InitAction,
    ReadFileAndAppVersion,
    init_file_and_app_version,
    init_state)
from analyst_klondike.features.data_context.set_opened_file_action import SetOpenedFileAction
from analyst_klondike.state.app_state import AppState
from analyst_klondike.state.base_action import BaseAction


def apply(state: AppState, action: BaseAction) -> AppState:
    if isinstance(action, InitAction):
        init_state(state, action.data)
        return state
    if isinstance(action, SetOpenedFileAction):
        state.current.opened_file_path = action.opened_file_path
        state.current.opened_file_name = action.opened_file_name
        return state
    if isinstance(action, ReadFileAndAppVersion):
        init_file_and_app_version(state, action.data)
    return state



import json
from typing import Any
import httpx
from analyst_klondike.features.events.event_sender import BaseEventSender


class ApiEventSender(BaseEventSender):

    def __init__(self, api_url: str) -> None:
        super().__init__()
        self.api_url = api_url

    async def send_event(self,
                         event_name: str,
                         event_location: str = "/",
                         *,
                         event_json: dict[str, str | int | bool] | None = None) -> None:
        try:
            await self._try_send_event(
                event_name, event_location, event_json=event_json
            )
        except Exception as ex:  # pylint: disable=broad-exception-caught
            print(f"fail to send event: {ex}")

    async def _try_send_event(self,
                              event_name: str,
                              event_location: str = "/",
                              *,
                              event_json: dict[str, str | int | bool] | None = None):
        user_ip = await self.get_user_ip()
        data_dict: dict[str, str | Any] = {
            "userId": user_ip,
            "eventJson": {} if event_json is None else {
                'addData': self._dict_to_kvps(event_json)
            },
            "eventName": event_name,
            "eventLocation": event_location,
            "eventLocationId": "",
            "userIp": user_ip
        }

        stringified_data = json.dumps(data_dict, ensure_ascii=False)

        async with httpx.AsyncClient(verify=False) as client:
            response = await client.post(
                url=self.api_url,
                content=stringified_data,
                timeout=5000,
                headers={
                    'accept': 'application/json',
                    'content-type': 'application/json'
                }
            )
            response.raise_for_status()

    def _dict_to_kvps(self, d: dict[str, str | int | bool] | str | None) -> str:
        if d is None:
            return ""
        if isinstance(d, str):
            return d
        return ", ".join(f"{k}={v}" for k, v in d.items())



from os import getcwd
from os.path import dirname, join, exists, basename

from analyst_klondike.features.code_import.ui_actions.actions import SwitchToTeacherModeAction
from analyst_klondike.features.data_context.set_opened_file_action import SetOpenedFileAction
from analyst_klondike.features.events.event_actions import DisableEvents
from analyst_klondike.state.app_dispatch import app_dispatch


class TooManyArgumentsException(Exception):
    pass


def process_args(argv: list[str]) -> None:
    if len(argv) < 2:
        return

    python_file_dir = dirname(argv[0])
    cmd = getcwd()
    fpath = _first_existed_path(argv[1:],
                                python_file_dir,
                                cmd)
    if "--no-events" in argv:
        app_dispatch(DisableEvents())
    if "--teacher" in argv:
        app_dispatch(SwitchToTeacherModeAction())
    if fpath is not None:
        fname = basename(fpath)
        app_dispatch(SetOpenedFileAction(
            opened_file_name=fname,
            opened_file_path=fpath
        ))


def _first_existed_path(file_names: list[str], *dirs: str) -> str | None:
    return next((
        join(d, f) for f in file_names for d in dirs if exists(join(d, f))
    ), None)



from analyst_klondike.features.events.event_sender import BaseEventSender


class NoEventSender(BaseEventSender):

    async def send_event(self,
                         event_name: str,
                         event_location: str = "/",
                         *,
                         event_json: dict[str, str | int | bool] | str | None = None):
        pass



# from analyst_klondike.features.code.code_runner.runner import CodeRunner


# _runner_cases: list[dict[str, object]] = [
#     {
#         "n": 1,
#         "expected": "1 просмотр"
#     },
#     {
#         "n": 2,
#         "expected": "2 просмотра"
#     },
#     {
#         "n": 2,
#         "expected": "2 просмотра"
#     },
#     {
#         "n": 3,
#         "expected": "3 просмотра"
#     },
#     {
#         "n": 10,
#         "expected": "10 просмотров"
#     },
#     {
#         "n": 21,
#         "expected": "21 просмотр"
#     }
# ]


# def test_run_code() -> None:
#     code = """\
# def solution(n: int) -> str:
#     view_dict = {
#         1: "1 просмотр",
#         2: "2 просмотра",
#         3: "3 просмотра",
#         4: "4 просмотра",
#     }
#     if n <= 4:
#         return view_dict[n]
#     return f"{n} просмотров"
#     """
#     result = CodeRunner().run_code(_runner_cases, code)
#     assert result is not None
#     assert len(list(result.all_cases)) > 0


# def test_overriding_print() -> None:
#     code = """\

# def solution(n: int) -> str:
#     view_dict = {
#         1: "1 просмотр",
#         2: "2 просмотра",
#         3: "3 просмотра",
#         4: "4 просмотра",
#     }
#     if n <= 4:
#         return view_dict[n]
#     print("hello")
#     print("world")
#     return f"{n} просмотров"
#     """
#     code_runner = CodeRunner()
#     result = code_runner.run_code(_runner_cases, code)
#     assert result is not None
#     assert "hello" in result.printed_lines
#     assert "world" in result.printed_lines


# def test_update_dict():
#     initial_dict = {
#         "a": 1,
#         "b": 2
#     }

#     initial_dict.update({
#         "c": 3,
#         "d": 4
#     })

#     assert "c" in initial_dict



import sys
from rich.live import Live
from rich.spinner import Spinner
from analyst_klondike.features.data_context.arg_file_load.arg_load import (
    TooManyArgumentsException,
    process_args
)
from analyst_klondike.ui.runner_app import get_app


def analyst_klondike():

    spinner = Spinner("dots", text="App is loading...")
    with Live(spinner, screen=True, refresh_per_second=10) as live:

        # clear console with spinner
        def _clear_console():
            live.console.clear()

        try:
            app = get_app()
            app.on_mounted_callback = _clear_console
            process_args(sys.argv)
            app.run()
        except (FileNotFoundError, TooManyArgumentsException) as err:
            live.console.print(err)
            sys.exit(1)


if __name__ == "__main__":
    analyst_klondike()



from dataclasses import dataclass
from analyst_klondike.state.app_state import AppState


def current_file_path(state: AppState) -> tuple[str, str]:
    return (
        state.current.opened_file_path,
        state.current.opened_file_name
    )


@dataclass
class VersionsInfo:
    app_version: str
    version_in_file: str
    is_file_compatible: bool


def versions(state: AppState) -> VersionsInfo:
    app_ver = state.current.app_version
    min_sup_ver = state.current.opened_file_min_supported_app_version

    return VersionsInfo(
        app_version=app_ver,
        version_in_file=min_sup_ver,
        is_file_compatible=is_compatible(app_ver, min_sup_ver)
    )


def is_compatible(app_ver: str, min_sup_ver: str) -> bool:
    av_tuple = app_ver.split(".")
    msp_tuple = min_sup_ver.split(".")
    return av_tuple >= msp_tuple



from dataclasses import dataclass
from typing import Any
import yaml

from textual.app import App

from analyst_klondike.features.data_context.data_state import TestCaseState
from analyst_klondike.features.message_box.actions import DisplayMessageBoxAction
from analyst_klondike.state.app_dispatch import app_dispatch
from analyst_klondike.state.app_state import AppState
from analyst_klondike.state.base_action import BaseAction


@dataclass
class SaveAction(BaseAction):
    type = "SAVE_ACTION"


def save_to_yaml(state: AppState, app: App[Any]) -> None:
    try:
        _try_save(state)
        app.notify(
            "Сохранено",
            title=state.current.opened_file_name,
            severity="information",
            timeout=1
        )
    except Exception:  # pylint: disable=W0718
        app_dispatch(DisplayMessageBoxAction(
            f"Ошибка при сохранении файла {state.current.opened_file_name}"
        ))
        app.notify(
            "Не удалось сохранить",
            title=state.current.opened_file_name,
            severity="error",
            timeout=1
        )


def _try_save(state: AppState) -> None:
    with open(state.current.opened_file_path,
              encoding='utf-8',
              mode='w') as f:
        data = _create_json(state)
        yaml.dump(data, f,
                  allow_unicode=True,
                  encoding='UTF-8',
                  default_flow_style=False,
                  indent=2,
                  sort_keys=False)


def _create_json(state: AppState) -> Any:
    def _get_tasks(quiz_id: str):
        task_dict = {
            t.title: t for t in state.data.tasks.values() if t.quiz_id == quiz_id
        }
        return task_dict.items()

    d = {  # type: ignore
        "quiz_info": {
            "min_supported_app_version": state.current.opened_file_min_supported_app_version
        },
        "user_info": {
            "email": state.user_email
        },
        "quizes": {
            quiz_id: {
                "title": quiz.title,
                "questions": {
                    t_id: {
                        "text": t.description,
                        "code_template": t.code_template,
                        "code": t.code,
                        "test_cases": _test_cases_to_json(t.test_cases),
                        "is_passed": t.is_passed,
                        "params": _get_task_params(t.test_cases)
                    } for t_id, t in _get_tasks(quiz.id)
                }
            } for quiz_id, quiz in state.data.quizes.items()
        }
    }
    return d  # type: ignore


def _get_task_params(cases: list[TestCaseState]) -> dict[str, str]:
    param_dict: dict[str, str] = {}

    for c in cases:
        for inp in c.inputs:
            param_dict[inp.param_name] = inp.param_type
        param_dict["return"] = c.expected.param_type
    return param_dict


def _test_cases_to_json(cases: list[TestCaseState]) -> list[dict[str, str]]:
    test_cases: list[dict[str, str]] = []
    for c in cases:
        test_cases_dict: dict[str, str] = {}
        for inp in c.inputs:
            test_cases_dict[inp.param_name] = inp.param_value
        test_cases_dict["expected"] = c.expected.param_value
        test_cases.append(test_cases_dict)
    return test_cases



from os.path import exists
from typing import Callable

from analyst_klondike.features.code.actions import RunAllCodeAction
from analyst_klondike.features.data_context.init_action import InitAction, ReadFileAndAppVersion
from analyst_klondike.features.data_context.json_load.dc import get_quiz_json
from analyst_klondike.features.data_context.selectors import current_file_path, versions
from analyst_klondike.features.message_box.actions import DisplayMessageBoxAction
from analyst_klondike.state.app_dispatch import app_dispatch
from analyst_klondike.state.app_state import select


def open_file(on_opened: Callable[[str | None], None]) -> None:
    # display message if file is not compatible
    fpath, fname = select(current_file_path)
    if not exists(fpath):
        return
    try:
        json_data = get_quiz_json(fpath)
    except Exception:  # pylint: disable=W0718
        app_dispatch(DisplayMessageBoxAction(
            f"Ошибка при открытии файла {fname}"
        ))
        return

    app_dispatch(ReadFileAndAppVersion(json_data))
    v_info = select(versions)
    if v_info.is_file_compatible:
        app_dispatch(InitAction(data=json_data))
        app_dispatch(RunAllCodeAction())
        on_opened(fname)
    else:
        app_dispatch(DisplayMessageBoxAction(
            f"""\
                    Файл можно будет открыть только в приложении версии {v_info.version_in_file}.
                    У вас установлена версия {v_info.app_version}.
                    Закройте приложение и обновите его, написав команду: 
                    [on blue]uv tool upgrade analyst-klondike[on blue]
                """
        ))
